import subprocess
import json
from shy_sh.settings import settings
from langchain_core.messages import HumanMessage, AIMessage, BaseMessage
from time import strftime
from rich import print
from rich.syntax import Syntax
from rich.live import Live
from shy_sh.agent.chains.shy_agent import shy_agent_chain
from shy_sh.agent.chains.python_expert import python_expert_chain
from shy_sh.agent.chains.shell_exec import shell_exec
from shy_sh.agent.chains.shell_expert import shell_expert_chain
from shy_sh.agent.chains.screenshot import screenshot_chain
from shy_sh.agent.utils import decode_output, detect_os, detect_shell

from .tools import tool
from .models import ToolRequest, FinalResponse


class ShyAgent:
    def __init__(
        self,
        max_iterations: int = 4,
        interactive=False,
        ask_before_execute=False,
        screenshot=False,
    ):
        self.max_iterations = max_iterations
        self.interactive = interactive
        self.ask_before_execute = ask_before_execute
        self.screenshot = screenshot
        self.history = []
        self.tools = self._get_tools()

    @property
    def formatted_tools(self):
        return "\n".join(
            map(lambda tool: f'- "{tool.name}": {tool.description}', self.tools)
        )

    def _get_tools(self):
        @tool
        def shell(arg: str):
            """to execute a shell command in the terminal, useful for every task that requires to interact with the current system or local files, do not pass interactive commands, avoid to install new packages if not explicitly requested"""
            return shell_exec(arg, self.ask_before_execute)

        @tool
        def python_expert(arg: str):
            """to delegate the task to a python expert that can write and execute python code, use only if you cant resolve the task with shell, just forward the task as argument without any python code"""
            return python_expert_chain(arg, self.history, self.ask_before_execute)

        @tool
        def shell_expert(arg: str):
            """to delegate the task to a shell expert that can write and execute complex shell scripts, use only if you cant resolve the task with a simple shell command, just forward the task as argument without any shell code"""
            return shell_expert_chain(arg, self.history, self.ask_before_execute)

        return [shell, shell_expert, python_expert]

    def _check_json(self, text: str):
        if text.count("{") > 0 and text.count("{") - text.count("}") == 0:
            first_bracket = text.index("{")
            last_bracket = text.rindex("}")
            maybe_tool = text[first_bracket : last_bracket + 1]
            if "mixtral" in settings.llm.name:
                maybe_tool = maybe_tool.replace("\\_", "_")  # fix mixtral bias
            try:
                return ToolRequest.model_validate_json(maybe_tool)
            except Exception:
                pass
        return text

    def _handle_tool(self, action: ToolRequest):
        tool = next((t for t in self.tools if t.name == action.tool), None)
        if tool is None:
            return f"Tool {action.tool} not found"
        try:
            tool_answer = tool.execute(action.arg)
        except Exception as e:
            tool_answer = f"🚨 There was an error: {e}"
        return tool_answer

    def _get_few_shot_examples(self):
        shell = detect_shell()
        os = detect_os()
        actions = [
            {
                "tool": "shell",
                "arg": "echo %cd%" if shell in ['powershell','cmd'] else "pwd",
                "thoughts": "I'm checking the current working directory",
            },
            {
                "tool": "shell",
                "arg": "git rev-parse --abbrev-ref HEAD",
                "thoughts": "I'm checking if it's a git repository",
            },
        ]
        result = []
        result.append(HumanMessage(content=f"You are on {os} system using {shell} as shell. Check your tools"))
        for action in actions:
            result.append(AIMessage(content=json.dumps(action)))
            response = subprocess.run(
                action["arg"],
                stdout=subprocess.PIPE,
                stderr=subprocess.PIPE,
                shell=True,
            )
            response = decode_output(response)

            result.append(HumanMessage(content=f"Tool response:\n{response}"))
        result.append(AIMessage(content="Ok"))
        return result

    def _update_task_with_image(self, task: str):
        result = screenshot_chain(task)
        return f"\nContext informations - This is what I'm seeing in my screen right now:\n{result}\n\nTask: {task}"

    def _execute(
        self,
        task: str,
        few_shot_examples: list[BaseMessage] = [],
    ):
        inputs = {
            "input": task,
            "timestamp": strftime("%Y-%m-%d %H:%M %Z"),
            "few_shot_examples": few_shot_examples,
            "history": self.history,
            "tool_history": self.tool_history,
            "tools": self.formatted_tools,
            "lang_spec": (
                f"translated in {settings.language} language "
                if settings.language
                else ""
            ),
        }

        stream_text = ""
        with Live() as live:
            loading = "⏱️  Loading..."
            live.update(loading)
            for chunk in shy_agent_chain.stream(inputs):
                stream_text += chunk
                if stream_text.startswith("{"):
                    live.update(loading)
                else:
                    live.update(f"🤖: {stream_text}", refresh=True)
                result = self._check_json(stream_text)
                if isinstance(result, ToolRequest):
                    live.update("")
                    self.tool_history.append(
                        AIMessage(content=result.model_dump_json())
                    )
                    result = self._handle_tool(result)
                    self.tool_history.append(
                        HumanMessage(content=f"Tool response:\n{result}")
                    )
                    return result
            live.update("")
        return FinalResponse(response=result)

    def _run(self, task: str, examples: list[BaseMessage] = []):
        answer = ""
        if self.screenshot:
            task = self._update_task_with_image(task)
            self.screenshot = False
        else:
            task = f"Task: {task}"
        for _ in range(self.max_iterations):
            answer = self._execute(task, examples)

            if isinstance(answer, FinalResponse):
                return answer.response
        return answer

    def start(self, task: str):
        self.tool_history = []
        examples = self._get_few_shot_examples()
        answer = None
        if task:
            answer = self._run(task, examples)
            print(
                Syntax(
                    f"🤖: {answer}",
                    "console",
                    theme="one-dark",
                    background_color="#181818",
                )
            )
        if self.interactive:
            new_task = input("\n✨: ")
            print()
            if new_task != "exit":
                self.history.append(HumanMessage(content=task))
                self.history += self.tool_history
                if answer:
                    self.history.append(AIMessage(content=answer))

                self.start(new_task)
            else:
                print("🤖: 👋 Bye!")



from typing import Literal
import platform
import os


def ask_confirm() -> Literal["y", "n", "c"]:
    return (
        input(
            "\n[dark_orange]Do you want to execute this command? [Y/n/c]:[/dark_orange] "
        )
        or "y"
    )[0].lower()


def decode_output(process):
    try:
        response = process.stdout.decode() or process.stderr.decode()
    except UnicodeDecodeError:
        # windows
        import ctypes
        oemCP = ctypes.windll.kernel32.GetConsoleOutputCP()
        encoding = 'cp'+str(oemCP)
        response = process.stdout.decode(encoding) or process.stderr.decode(encoding)
    return response

def detect_shell():
    shell = os.environ.get('SHELL') or os.environ.get('COMSPEC')
    shell = shell.lower()
    if 'powershell' in shell:
        return 'powershell'
    elif 'cmd' in shell:
        return 'cmd'
    
    return shell.lower()

def detect_os():
    return platform.system()


from langchain_core.prompts import ChatPromptTemplate, MessagesPlaceholder
from langchain_core.output_parsers import StrOutputParser
from langchain_core.runnables import chain
from shy_sh.agent.chat_models import get_llm
from textwrap import dedent

sys_template = dedent(
    """
    You are a helpfull shell assistant. The current date and time is {timestamp}.
    Try to resolve the tasks that I request you to do.

    You can use the following tools to accomplish the tasks:
    {tools}
                        
    Rules:
    You can use only the tools provided in this prompt to accomplish the tasks
    If you need to use tools your response must be in JSON format with this structure: {{ "tool": "...", "arg": "...", "thoughts": "..." }}
    Use the shell and your other tools to gather all the information that you need before starting the actual task and also to double check the results if needed before giving the final answer
    After you completed the task output your final answer to the task {lang_spec}without including any json
    Answer truthfully with the informations you have
    You cannot use tools and complete the task with your final answer in the same message so remember to use the tools that you need first
    """
)


@chain
def shy_agent_chain(_):
    llm = get_llm()
    prompt = ChatPromptTemplate.from_messages(
        [
            ("system", sys_template),
            MessagesPlaceholder("few_shot_examples", optional=True),
            MessagesPlaceholder("history", optional=True),
            ("human", "{input}"),
            MessagesPlaceholder("tool_history", optional=True),
        ]
    )
    return prompt | llm | StrOutputParser()



import typer
from typing import Optional, Annotated
from importlib.metadata import version
from shy_sh.agent.agent import ShyAgent
from shy_sh.settings import settings, configure_yaml
from rich import print


def exec(
    prompt: Annotated[Optional[list[str]], typer.Argument(allow_dash=False)] = None,
    interactive: Annotated[
        Optional[bool],
        typer.Option(
            "-i",
            help="Interactive mode [default false if a prompt is passed]",
        ),
    ] = False,
    no_ask: Annotated[
        Optional[bool],
        typer.Option(
            "-x",
            help="Do not ask for confirmation before executing scripts",
        ),
    ] = False,
    screenshot: Annotated[
        Optional[bool],
        typer.Option(
            "-s",
            help="Take a screenshot of the terminal before the execution",
        ),
    ] = False,
    configure: Annotated[
        Optional[bool], typer.Option("--configure", help="Configure LLM")
    ] = False,
    display_version: Annotated[
        Optional[bool], typer.Option("--version", help="Show version")
    ] = False,
):
    if display_version:
        print(f"Version: {version(__package__ or 'shy-sh')}")
        return
    if configure:
        configure_yaml()
        return
    task = " ".join(prompt or [])
    print(f"[bold italic dark_orange]{settings.llm.provider} - {settings.llm.name}[/]")
    if not task:
        interactive = True
    else:
        print(f"✨: {task}\n")
    try:
        ShyAgent(
            interactive=interactive,
            ask_before_execute=not no_ask,
            screenshot=screenshot,
        ).start(task)
    except Exception as e:
        print(f"🚨 [bold red]{e}[/bold red]")


def main():
    typer.run(exec)


if __name__ == "__main__":
    main()



from typing import Callable
from abc import ABC, abstractmethod


class Tool(ABC):
    def __init__(
        self, name: str | None = None, description: str | None = None, **kwargs
    ):
        self.name = name or self.__class__.__name__
        self.description = description or self.__doc__
        self.kwargs = kwargs

    @abstractmethod
    def execute(self, payload: str): ...


def tool(*args, **kwargs) -> Tool:
    def decorator(fn: Callable):
        class ToolImpl(Tool):
            def execute(self, payload: str):
                return fn(payload)

        return ToolImpl(fn.__name__, fn.__doc__, **kwargs)

    if args:
        return decorator(args[0])
    return decorator



from langchain_core.messages import HumanMessage
from langchain_core.output_parsers import StrOutputParser
from shy_sh.agent.chat_models import get_llm
from rich import print
from base64 import b64encode
from PIL import ImageGrab
from io import BytesIO


def screenshot_chain(task: str):
    print(f"📸 [bold yellow]Taking a screenshot...[/bold yellow]")
    image = ImageGrab.grab().convert("RGB")
    buffered = BytesIO()
    image.save(buffered, format="JPEG")
    image_b64 = b64encode(buffered.getvalue()).decode("utf-8")

    llm = get_llm()
    message = HumanMessage(
        content=[
            {
                "type": "text",
                "text": f"Write a concise descrption of the image.\nDescribe what you see but only the parts that are usefull to solve this task: {task}",
            },
            {
                "type": "image_url",
                "image_url": {"url": f"data:image/jpeg;base64,{image_b64}"},
            },
        ],
    )
    result = ""
    print(f"🤖: ", end="")
    chain = llm | StrOutputParser()
    for chunk in chain.stream([message]):
        print(chunk, end="", flush=True)
        result += chunk
    print()
    return result



import pyperclip
from io import StringIO
from contextlib import redirect_stdout
from time import strftime
from rich.syntax import Syntax
from shy_sh.agent.models import FinalResponse
from langchain_core.runnables import chain
from langchain_core.output_parsers import StrOutputParser
from langchain.prompts import ChatPromptTemplate, MessagesPlaceholder
from shy_sh.agent.chat_models import get_llm
from shy_sh.agent.utils import ask_confirm
from textwrap import dedent
from rich import print

sys_template = dedent(
    """
    Output only a block of python code like this:
    ```python
    [your python code]
    ```

    Do not wrap your script in if __name__ == "__main__": block

    Write a python script that accomplishes the task.
    Task: {input}
    """
)


@chain
def _chain(_):
    llm = get_llm()
    prompt = ChatPromptTemplate.from_messages(
        [
            (
                "system",
                "You are a python expert. The current date and time is {timestamp}",
            ),
            MessagesPlaceholder("history", optional=True),
            ("human", sys_template),
        ]
    )
    return prompt | llm | StrOutputParser()


def python_expert_chain(task: str, history, ask_before_execute: bool):
    print(f"🐍 [bold yellow]Generating python script...[/bold yellow]\n")
    code = _chain.invoke(
        {
            "input": task,
            "timestamp": strftime("%Y-%m-%d %H:%M %Z"),
            "history": history,
        }
    )
    code = code.replace("```python\n", "").replace("```", "")
    print(Syntax(code.strip(), "python", background_color="#212121"))
    if ask_before_execute:
        confirm = ask_confirm()
        if confirm == "n":
            return FinalResponse(response="Command canceled by user")
        elif confirm == "c":
            pyperclip.copy(code)
            return FinalResponse(response="Script copied to the clipboard!")
    stdout = StringIO()
    with redirect_stdout(stdout):
        exec(code)
    output = stdout.getvalue().strip() or "Done"
    print(Syntax(output, "python", background_color="#212121"))
    return f"\nScript executed:\n```python\n{code.strip()}\n```\n\nOutput:\n{output}"



from shy_sh.settings import settings
from functools import lru_cache


@lru_cache
def get_llm():
    llm = None
    match settings.llm.provider:
        case "openai":
            from langchain.chat_models.openai import ChatOpenAI

            llm = ChatOpenAI(
                model=settings.llm.name,
                temperature=settings.llm.temperature,
                api_key=settings.llm.api_key,
            )
        case "ollama":
            from langchain_community.chat_models.ollama import ChatOllama

            llm = ChatOllama(
                model=settings.llm.name, temperature=settings.llm.temperature
            )

        case "groq":
            from langchain_groq import ChatGroq

            llm = ChatGroq(
                model=settings.llm.name,
                temperature=settings.llm.temperature,
                api_key=settings.llm.api_key,
            )

        case "anthropic":
            from langchain_community.chat_models.anthropic import ChatAnthropic

            llm = ChatAnthropic(
                model_name=settings.llm.name,
                temperature=settings.llm.temperature,
                anthropic_api_key=settings.llm.api_key,
            )

        case "google":
            from langchain_google_genai import ChatGoogleGenerativeAI
            import os

            # Suppress logging warnings
            os.environ["GRPC_VERBOSITY"] = "ERROR"
            os.environ["GLOG_minloglevel"] = "2"

            llm = ChatGoogleGenerativeAI(
                model=settings.llm.name,
                temperature=settings.llm.temperature,
                api_key=settings.llm.api_key,
            )

        case "aws":
            from langchain_aws import ChatBedrock
            import boto3

            region, access_key, secret_key = settings.llm.api_key.split(" ")

            boto_client = boto3.client(
                "bedrock-runtime",
                aws_access_key_id=access_key,
                aws_secret_access_key=secret_key,
                region_name=region,
            )

            llm = ChatBedrock(
                model_id=settings.llm.name,
                client=boto_client,
                model_kwargs={"temperature": settings.llm.temperature},
            )
    return llm



import os
import yaml

from typing import Type, Any

from pydantic_settings import (
    BaseSettings,
    SettingsConfigDict,
    YamlConfigSettingsSource,
)
from pydantic import BaseModel
from pathlib import Path
from rich.prompt import Prompt, FloatPrompt


class LLMSchema(BaseModel):
    provider: str
    name: str
    api_key: str = ""
    temperature: float = 0.0


class Settings(BaseSettings):
    model_config = SettingsConfigDict(
        yaml_file=[
            "~/.config/shy/config.yml",
            "~/.config/shy/config.yaml",
            "./shy.yml",
            "./shy.yaml",
        ]
    )

    llm: LLMSchema = LLMSchema(provider="ollama", name="llama3.1")

    language: str = ""

    @classmethod
    def settings_customise_sources(
        cls,
        settings_cls: Type[BaseSettings],
        **kwargs: Any,
    ):
        return (YamlConfigSettingsSource(settings_cls),)


settings = Settings()
PROVIDERS = ["ollama", "openai", "google", "anthropic", "groq", "aws"]


def get_or_create_settings_path():
    file_name = None
    for f in settings.model_config["yaml_file"]:
        f = Path(f).expanduser()
        if os.path.exists(f):
            file_name = f
            break
    if not file_name:
        file_name = settings.model_config["yaml_file"][0]
        file_name = Path(file_name).expanduser()
        os.makedirs(os.path.dirname(file_name), exist_ok=True)
    return file_name


def configure_yaml():
    provider = Prompt.ask(
        "[green bold]Provider[/]",
        default=settings.llm.provider,
        show_choices=True,
        choices=PROVIDERS,
    )
    name = Prompt.ask(
        "[green bold]Model[/]",
        default=settings.llm.name,
    )

    api_key = Prompt.ask(
        "[green bold]API Key[/]",
        default=settings.llm.api_key,
        password=True,
        show_default=False,
    )
    temperature = FloatPrompt.ask(
        "[green bold]Temperature[/]",
        default=settings.llm.temperature,
    )
    language = Prompt.ask(
        "[green bold]Language[/]",
        default=settings.language,
    )

    file_name = get_or_create_settings_path()

    with open(file_name, "w") as f:
        f.write(
            yaml.dump(
                {
                    "llm": {
                        "provider": provider,
                        "name": name,
                        "api_key": api_key,
                        "temperature": temperature,
                    },
                    "language": language,
                }
            )
        )

    print(f"Configuration saved to {file_name}")



from typing import Optional
from pydantic import BaseModel


class ToolRequest(BaseModel):
    tool: str
    arg: str
    thoughts: Optional[str] = None


class FinalResponse(BaseModel):
    response: str



import os
import subprocess
import pyperclip
from time import strftime
from rich.syntax import Syntax
from shy_sh.agent.models import FinalResponse
from langchain_core.runnables import chain
from langchain_core.output_parsers import StrOutputParser
from langchain.prompts import ChatPromptTemplate, MessagesPlaceholder
from shy_sh.agent.chat_models import get_llm
from shy_sh.agent.utils import ask_confirm, decode_output, detect_shell,detect_os
from textwrap import dedent
from rich import print
from uuid import uuid4


sys_template = dedent(
    """
    Output only a block of code like this:
    ```sh
    #!/bin/sh
    set -e
    [your shell script]
    ```

    This is a template for sh, but you should use the right shell syntaxt depending on your system.
    Don't write interactive commands or install new packages if not explicitly requested.
    Write a shell script that accomplishes the task.

    Task: {input}
    """
)


@chain
def _chain(_):
    llm = get_llm()
    prompt = ChatPromptTemplate.from_messages(
        [
            (
                "system",
                "You are a shell expert. The current date and time is {timestamp}\nYou are running on {system} using {shell} as shell",
            ),
            MessagesPlaceholder("history", optional=True),
            ("human", sys_template),
        ]
    )
    return prompt | llm | StrOutputParser()


def shell_expert_chain(task: str, history, ask_before_execute: bool):
    print(f"👨‍💻 [bold yellow]Generating shell script...[/bold yellow]\n{task}\n\n")
    shell = detect_shell()
    system = detect_os()
    code = _chain.invoke(
        {
            "input": task,
            "timestamp": strftime("%Y-%m-%d %H:%M %Z"),
            "history": history,
            "system": system,
            "shell": shell
        }
    )
    code = code.replace("```sh\n", "").replace("```", "")
    print(Syntax(code.strip(), "console", background_color="#212121"))
    if ask_before_execute:
        confirm = ask_confirm()
        if confirm == "n":
            return FinalResponse(response="Task interrupted")
        elif confirm == "c":
            pyperclip.copy(code)
            return FinalResponse(response="Script copied to the clipboard!")

    file_path = f"/tmp/{uuid4().hex}.sh"
    with open(file_path, "w") as f:
        f.write(code)
    os.chmod(file_path, 0o755)
    result = subprocess.run(
        file_path,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,
    )
    stdout = decode_output(result) or "Done"
    print(Syntax(stdout.strip(), "console", background_color="#212121"))
    return f"Script executed:\n{code}\n\nOutput:\n{stdout}"



import subprocess
import pyperclip
from rich.syntax import Syntax
from shy_sh.agent.models import FinalResponse
from shy_sh.agent.utils import ask_confirm, decode_output
from rich import print


def shell_exec(arg: str, ask_before_execute: bool):
    print(f"🛠️ [bold green]{arg}[/bold green]")
    if ask_before_execute:
        confirm = ask_confirm()
        if confirm == "n":
            return FinalResponse(response="Task interrupted")
        elif confirm == "c":
            pyperclip.copy(arg)
            return FinalResponse(response="Command copied to the clipboard!")

    result = subprocess.run(
        arg,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        shell=True,
    )
    stdout = decode_output(result) or "Success!"
    print(Syntax(stdout.strip(), "console", background_color="#212121"))
    return stdout
